/* Implements a camera reference system with some transformations and
camera data.
Example:
             J
            |
            |  / K
            | /
            |/
            *------
             O     I

TIPS:
- this class is derived from CReferenceSystem, see refsys.cc for more
  information.

WARNING:
- after change floating point base of the reference system you must update
  the fix point base by calling CCameraSystem::mComputeFixPointBase.
*/



#include <string.h>

#include <math.h>

#include "defs.h"

#include "node.h"

#include "list.h"

#include "polygon.h"

#include "nodepol.h"

#include "nodever.h"

#include "nodexy.h"

#include "math.h"

#include "refsys.h"

#include "camsys.h"



/* constructor - init default reference system
**********************************************************/
CCameraSystem::CCameraSystem ()
{
   mCreate ();
}



/* constructor - init reference system with other one.
**********************************************************/
CCameraSystem::CCameraSystem ( CCameraSystem& oCS )
{
   mCreate ();
   mSetObj ( oCS );
}



/* destructor
**********************************************************/
CCameraSystem::~CCameraSystem ()
{
}



/* Init reference system and lockup tables
**********************************************************/
void CCameraSystem::mCreate ()
{
   CReferenceSystem::mCreate ();

   m_nDistCP = DIST_CP;
   m_fRadius = RAD;   
   m_fSpeed = SPEED;
}



/* Destroy object
**********************************************************/
void CCameraSystem::mDestroy ()
{
   CReferenceSystem::mDestroy ();
}



/* Init reference system with a copy of other object
(like CCameraSystem::constructor and CCameraSystem::operator =)
**********************************************************/
void CCameraSystem::mSetObj ( CCameraSystem& oCS )
{
   CReferenceSystem::mSetObj ( oCS );

   // distance of projection center and projection plane
   m_nDistCP = oCS.m_nDistCP;
   // radius
   m_fRadius = oCS.m_fRadius;
   // move speed
   m_fSpeed = oCS.m_fSpeed;
}



/* like CCameraSystem::constructor and CCameraSystem::mSetObj
**********************************************************/
void CCameraSystem::operator = ( CCameraSystem& oCS )
{
   mSetObj ( oCS );
}
