/* Basic definitions */



// verify if has been declared, if no do it
#ifndef __3DFC_DEFS_H__



// signals file inclusion
#define __3DFC_DEFS_H__



// define compilation platform target and
// other compilation features
#include "../config.h"



// TYPES

// unsigned int 8 bits
#define BYTE unsigned char

// int (signed and unsigned) 32 bits
#define INT  int
#define UINT unsigned int
#define DWORD unsigned int

// int (signed and unsigned) 16 bits
#define INT16  short int
#define UINT16 unsigned short int
#define WORD unsigned short int

// pointer to uint (portability to 64 bits systems - pointers have 64 bits)
#define PTRVAL unsigned int

// boolean
#define BOOL unsigned char
#define TRUE 1
#define FALSE 0



// STRUCTURES

// describe a start and end value for an interval
struct tagInterval {
   INT
      nStart,
      nEnd;
};

// describe a location in 2D plane, coordenate has only 2 components: x and y
struct tagXY {
   INT
      nX,
      nY;
};

// describe a location in 3D space, coordenate has 3 components: x, y, z
struct tagXYZ {
   INT
      nX,
      nY,
      nZ;
};

// describe a location in 3D space, coordenate has 3 components: x, y, z
// this version use float
struct tagfXYZ {
   float
      fX,
      fY,
      fZ;
};

// define a rectangle
struct tagRectangle {
   INT nStartX;
   INT nStartY;
   INT nEndX;
   INT nEndY;
};

// define the destination of a portal polygon: the object and sector IDs
struct tagPortal {
   DWORD dwObjectID;     // ID for object where destin sector is
   DWORD dwSectorID;     // destin sector ID
};



// MESSAGES

#define MSG_FILE_NOT_FOUND "\nERROR: file '%s' not found or error reading.\n", (char*) oszFile
#define MSG_DEFAULT_ABORT_INFO "(problem in '%s', line %d)\n"
#define MSG_ERROR_MEMORY_LOW "\nERROR: impossible alocate memory, free memory and try again.\n"
#define MSG_USER_ABORT "\nProgram aborted by user.\n"
#define MSG_ERROR_READING_TOKEN "\nERROR reading token, possible end of file found.\n"
#define MSG_ERROR_READING_FILE "\nERROR: reading file '%s'.\n", (char*) m_oszFile
#define MSG_ERROR_REMARK_OPENED "\nERROR: reading file '%s', remark still open in level %d.\n", (char*) m_oszFile, m_nRemarkLevel
#define MSG_WARN_END_REMARK "\nWarning: an end remark simbol found without start remark simbol.\n"
#define MSG_ERROR_CREATING_GR "\nError when alocating MCGA graphics class instance.\n"
#define MSG_ERROR_BLOCK "ERROR: unknow block `%s' found in `%s' block.\n"
#define MSG_NOBLOCK "\nERROR: no block found `%s'in `%s' block, BUGS?!\n"
#define MSG_TOKEN_FOUND "Found: `%s' in file '%s'.\n"
#define MSG_ERROR_SPECTED "ERROR: not found %s after `%s'.\n"
#define MSG_STR_FILE_NAME "file name"
#define MSG_NOT_FOUND "ERROR: not found '%s'.\n"
#define MSG_STR_ID "identifier"
#define MSG_READING_TEXTURE "Reading texture from PCX file '%s'.\n"
#define MSG_READING_LIGHTTABLE "Reading light table from LTB file '%s'.\n"
#define MSG_READING_TXE "\nReading TXE file '%s'.\n"
#define MSG_END_READING_TXE "End of reading TXE file '%s'.\n"
#define MSG_READING_PALETTE "Reading palette from PCX file '%s'.\n"
#define MSG_CONTINUE_READING_TXE "\nContinue reading TXE file '%s'...\n"
#define MSG_STR_INSTANCES "scenery instances"
#define MSG_STR_VALUE "numeric value"
#define MSG_STR_OBJECT_BLOCK "block declaration inside object"
#define MSG_ERROR_SPECTED_IN "ERROR: spected `%s' in `%s' block.\n"
#define MSG_ERROR_UNSPECTED_EOF "ERROR: unspected end-of-file found in file '%s', block type `%s' still open.\n"
#define MSG_ERROR_UNSPECTED_EOB "ERROR: unspected close block found in file '%s', block `%s'.\n"
#define MSG_ERROR_READING_BLOCK "Error reading block `%s' identified by `%d'.\n"
#define MSG_ERROR_PALETTE_ALREADY_DEFINED "ERROR: palette already defined.\n"
#define MSG_ERROR_TEXTURE_ID_ALREADY_DEFINED "ERROR: texture ID `%d' already defined.\n"
#define MSG_ERROR_OBJECT_ID_ALREADY_DEFINED "ERROR: object ID `%d' already defined.\n"
#define MSG_ERROR_SECTOR_ID_ALREADY_DEFINED "ERROR: in object with ID `%d' the sector ID `%d' already defined.\n"
#define MSG_ERROR_LIGHTTABLE_ID_ALREADY_DEFINED "ERROR: light table ID `%d' already defined.\n"
#define MSG_ERROR_NO_LIGHTTABLE "ERROR: no light table defined with this ID.\n"
#define MSG_ERROR_NOPALETTE_FOR_LIGHTTABLE "ERROR: light table file not found and impossible create light table. No palette defined, load palette first.\n"
#define MSG_ERROR_INDEX_OUT_OF_RANGE "ERROR: index out of range, try correct declaring all polygons first and then their attributes.\n"
#define MSG_ERROR_ID_NOT_FOUND "ERROR: undefined ID.\n"
#define MSG_ERROR_TRAINING_VECTORS_UNDEFINIED "ERROR: define first one the training vector.\n"
#define MSG_ERROR_LOW_VERTEXES "ERROR: not enough vertexes in polygon (minimum 3).\n"
#define MSG_ERROR_MANY_VERTEXES "ERROR: to much vertexes in polygon.\n"
#define MSG_ERROR_NO_TEXTURE "ERROR: texture ID not found.\n"
#define MSG_FRAMES "\nTime: %f secs.\nFrames: %u.\nFrames per sec: %f\n"
#define MSG_ERROR_OVER_RESY_LIMIT "ERROR: to much resolution in Y frame buffer.\n"
#define MSG_ERROR_SECTOR_NOT_FOUND "ERROR: sector not found.\n"
#define MSG_ERROR_OBJECT_NOT_FOUND "ERROR: object not found.\n"
#define MSG_ERROR_LOW_VERTEXES_IN_MESH "ERROR: not enough vertexes defined in mesh (minimum 3).\n"
#define MSG_ERROR_SOCKET_CREATE "ERROR: impossible create socket.\n"
#define MSG_ERROR_SOCKET_BIND_NAME "ERROR: impossible bind a name to socket.\n"
#define MSG_ERROR_NO_CAMERA "ERROR: no camera with ID 0 defined.\n"
#define MSG_ERROR_CREATING_SOCKET "RROR: impossible create socket.\n"
#define MSG_ERROR_DESTIN_OBJECT_NOT_FOUND "ERROR: destin object instance '%d' not found.\n"
#define MSG_ERROR_DESTIN_SECTOR_NOT_FOUND "ERROR: destin instance sector '%d' not found.\n"
#define MSG_ERROR_SOURCE_MODEL_NOT_FOUND "ERROR: source object model '%d' not found when making instanciation.\n"
#define MSG_ERROR_CAMERA_OBJECTID_NOT_FOUND "ERROR: object `%d' not found, changing current camera.\n"
#define MSG_ERROR_SOCKET_CREATION "ERROR: impossible create socket.\n"
#define MSG_OBJECT_ID "Object ID %d.\n"
#define MSG_SECTOR_ID "Sector ID %d.\n"


// FILE SYSTEM

// reserved simbols
#define STR_READ "Read"
#define STR_TEXTURES "Textures"
#define STR_PALETTE "Palette"
#define STR_OBJECT "Object"
#define STR_SECTOR "Sector"
#define STR_POLYGONS "Polygons"
#define STR_VERTEXES "Vertexes"
#define STR_PORTALS "Portals"
#define STR_ATTRIBUTES "Attributes"
#define STR_ERROR "unknow-block-error"
#define STR_NOBLOCK "top-outside-all-blocks"
#define STR_DATA "simple data"
#define STR_OPENBLOCK "{"
#define STR_CLOSEBLOCK "}"
#define STR_OPENTUPLE "("
#define STR_CLOSETUPLE ")"
#define STR_INSTANCES "Instances"
#define STR_START "Start"
#define STR_LIGHTTABLES "LightTables"
#define STR_POLYGONSLIGHT "PolygonsLight"

#define STR_TUPLESEPARATOR ";"
#define STR_CONTINUATOR ".."

// id codes (used to define blocks types)
// ID_NOBLOCK is the topmost and ID_ERROR is a block not reconized (error)
// ID_DATA is a data found in the right place (no reserved word)

#define NUM_IDS 20

#define ID_ERROR 0
#define ID_NOBLOCK 1
#define ID_DATA 2
#define ID_READ 3
#define ID_TEXTURES 4
#define ID_PALETTE 5
#define ID_OBJECT 6
#define ID_SECTOR 7
#define ID_POLYGONS 8
#define ID_VERTEXES 9
#define ID_PORTALS 10
#define ID_ATTRIBUTES 11
#define ID_OPENBLOCK 12
#define ID_CLOSEBLOCK 13
#define ID_OPENTUPLE 14
#define ID_CLOSETUPLE 15
#define ID_INSTANCES 16
#define ID_START 17
#define ID_LIGHTTABLES 18
#define ID_POLYGONSLIGHT 19



// SYSTEM SPECIFIC

// path and file name separator '/' or '\'
#define _3DFC_FILESEPARATOR '/'
#ifdef TRG_DJGPP
   #define _SYSTEM_FILESEPARATOR '\\'
#endif
#ifdef TRG_LINUX
   #define _SYSTEM_FILESEPARATOR '/'
#endif

// drive separator
#define DRIVESEPARATOR ':'



// VIDEO MODES

#define MODE_TEXT 0
#define MODE_MCGA 1



// REFERENCE SYSTEM AND ENGINE

// Limits:

// how many units perform one round
#define ROUND 256

// maximum Y resolution for frame buffers
#define MAX_RESY 256

// number of colors in palette
#define MAXCOLS 256


// Metrics and scales:

// Z clip plan position
#define Z_CLIP 0

// distance between projection plan and projection center
#define DIST_CP -350

// camera radius
#define RAD 400

// maximum speed
#define SPEED 70

// scale factor for objects (position is position * 2^SCALEEXP)
#define SCALEEXP 3


// Lights and colors:

// signature for light table file
#define LIGHT_TABLE_SIGN "LT"

// version control for light table file
#define LIGHT_TABLE_VERSION 0

// palette range for each component: r, g, b
#define PAL_RANGE 64

// default light level
#ifndef COMP_COLOR_LIGHTS
#define DEFAULT_LIGHT_LEVEL 1
#endif
#ifdef COMP_COLOR_LIGHTS
#define DEFAULT_LIGHT_LEVEL 0.5
#endif


// Texture:

// texture size
#define TEXTURE_SIZE 256

// texture max dimension (TEXTURE_SIZE-1)
#define TEXTURE_MAXDIM 255

// values used in iner loop in poly line with textures
#define LINE_SHIFT_X 16
#define LINE_SHIFT_Y 8
#define LINE_AND_X TEXTURE_MAXDIM
#define LINE_AND_Y 0xff00



// GERENAL USE MACROS

// verify if files header needed have been already included, if no include
#ifdef TRG_DJGPP
   #ifndef __dj_include_stdio_h_
      #include <stdio.h>
   #endif

   #ifndef __dj_include_conio_h_
      #include <conio.h>
   #endif

   #ifndef __dj_include_stdlib_h_
      #include <stdlib.h>
   #endif

   #ifndef __dj_include_dos_h_
      #include <dos.h>
   #endif

   #include "djgpp\\keyboard.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#ifdef TRG_LINUX
   #include <vga.h>     // vga graphics library for Linux   :)
   #include <sys/types.h>
   #include <sys/socket.h>
   #include <netinet/in.h>
   #include <netdb.h>
   #include <errno.h>
#endif

// Abort program execution:
// set text mode, write message and terminate
#ifdef TRG_DJGPP
   #define ABORT(MSG,FILE,LINE) \
   { \
      REGS \
         stRegsE, stRegsS; \
      stRegsE.x.ax = 0x0003; \
      int86 ( 0x10, &stRegsE, &stRegsS ); \
      printf ( MSG ); \
      printf ( MSG_DEFAULT_ABORT_INFO, FILE, LINE ); \
      exit ( 1 ); \
   }
#endif
#ifdef TRG_LINUX
   #define ABORT(MSG,FILE,LINE) \
   { \
      vga_setmode ( TEXT ); \
      printf ( MSG ); \
      printf ( MSG_DEFAULT_ABORT_INFO, FILE, LINE ); \
      exit ( 1 ); \
   }
#endif

// Set pixel directily on video memory (only work in MCGA mode)
#ifdef TRG_DJGPP
   #define SETPIXEL(X,Y,C) \
   { \
      BYTE \
         yCol = C; \
      if ( ( X ) >= 0 && ( X ) < 320 && ( Y ) >= 0 && ( Y ) < 200 ) \
         dosmemput ( (void*) &yCol, 1, 0xA0000 + ( X ) + ( ( Y ) * 320 ) ); \
   }
#endif



// CLASSES PREDECLARATIONS

class CFilePCX;
class CTexture;

class CFileTXE;
class CScenery;

class CNodeSector;
class CNodeObject;



// end declarations
#endif
