/* Read a text file that describe scenery elements (objects, sectors, define
textures palettes, ...). The data loaded are stored in a CScenery instance.

The file extension commonly used for this files is TXE (TeXt + Elements).

Use the UNIX file separator (/) to refer files when possible to make all more
compatible with UNIX and DOS.
*/



#include <stdlib.h>

#include "defs.h"

#include "mem.h"

#include "string.h"

#include "node.h"

#include "list.h"

#include "polygon.h"

#include "nodepol.h"

#include "nodeint.h"

#include "file.h"

#include "filerem.h"

#include "texture.h"

#include "lighttab.h"

#include "filepcx.h"

#include "nodetex.h"

#include "map.h"

#include "mesh.h"

#include "sector.h"

#include "nodesec.h"

#include "math.h"

#include "refsys.h"

#include "camsys.h"

#include "scenery.h"

#include "filetxe.h"



/* constructor - define no file to read
**********************************************************/
CFileTXE::CFileTXE ()
{
   m_poScen = NULL;
   mDefineIDNames ();
}



/* constructor - read object from file, if fail abort.
**********************************************************/
CFileTXE::CFileTXE ( CString oszFile, CScenery& oScenery )
{
   m_poScen = NULL;
   mDefineIDNames ();

   // try load file
   if ( ! mLoad ( oszFile, oScenery ) )
      ABORT ( MSG_FILE_NOT_FOUND, __FILE__, __LINE__ )
}



/* destructor
**********************************************************/
CFileTXE::~CFileTXE ()
{
   m_poScen = NULL;
   mClose ();
}



/* Load data from a TXE file to a instance of CScenery.
oszFile - file name (with path and extension)
oScenery - where store data
return TRUE - ok / FALSE - error
**********************************************************/
BOOL CFileTXE::mLoad ( CString oszFile, CScenery& oScenery )
{
   m_poScen = &oScenery;

   // try open file
   if ( ! mOpen ( oszFile, "rt" ) ) {
      printf ( MSG_FILE_NOT_FOUND );
      return FALSE;
   }

   printf ( MSG_READING_TXE, (char*) oszFile );

   // read TXE file blocks (see ID values in header file)
   INT
      nReturnCode = mReadBlock ( oszFile, ID_NOBLOCK );

   printf ( MSG_END_READING_TXE, (char*) oszFile );

   // end load
   mClose ();
   m_poScen = NULL;
   if ( nReturnCode == 1 )
      return TRUE;
   else
      return FALSE;
}



/* Read the beginig of TXE file blocks and define where to
store data in CScenery object.
oszFile - file beeing read,
nBlockType - type of block parent of this block
return 1 if ok else error
**********************************************************/
INT CFileTXE::mReadBlock ( CString oszFile, INT nParentBlock )
{
   // while not end do read
   BOOL
      bEnd = FALSE;
   while ( ! bEnd ) {

      // next token string to be read
      CString
         oszStr;

      // try read token, if error supose end of file
      if ( ! mRetStr ( oszStr ) )
      
         // verify if now read cursor is in top-outside-every-blocks
         if ( nParentBlock == ID_NOBLOCK )
            // if no block it's ok find end-of-file
            break;
         else {  // if inside one block it's wrong find end-of-file
            printf ( MSG_ERROR_UNSPECTED_EOF, (char*) oszFile, (char*) m_aoIDs [ nParentBlock ] );
            return 0;
         }

      // show what has been found
      printf ( MSG_TOKEN_FOUND, (char*) oszStr, (char*) oszFile );

      // define type of block (open block is one possible name!)
      switch ( mClassifyBlock ( oszStr, nParentBlock ) ) {

         case ID_ERROR:
            // REVIEW LINUX problema para achar fim arquivo?
            if ( nParentBlock != ID_NOBLOCK ) {
               printf ( MSG_ERROR_BLOCK, (char*) oszStr, (char*) m_aoIDs [ nParentBlock ] );
               return 0;
            }
            else
               bEnd = TRUE;
            break;

         case ID_NOBLOCK:
            printf ( MSG_NOBLOCK, (char*) oszStr, (char*) m_aoIDs [ nParentBlock ] );
            return 0;
            break;

         case ID_READ:
            // try read open block, if can't return
            if ( ! mTestNextToken ( STR_OPENBLOCK ) ) {
               printf ( MSG_ERROR_SPECTED, STR_OPENBLOCK, STR_READ );
               return 0;
            }

            // read file names and open them (one at a time)
            do {

               // try read token, if error: exit
               if ( ! mRetStr ( oszStr ) ) {
                  printf ( MSG_ERROR_SPECTED, MSG_STR_FILE_NAME, STR_READ );
                  return 0;
               }

               // convert 3DFC file separator '/' to system file seperator
               oszStr.mReplace ( _3DFC_FILESEPARATOR, _SYSTEM_FILESEPARATOR );

               // verify if not found end of block
               if ( oszStr != STR_CLOSEBLOCK ) {

                  // read file
                  // define file name to be readed
                  CString
                     oszPath;
                    // verify if the name readed not have drive-separator
                  if ( ! oszStr.mTestFromLeft ( DRIVESEPARATOR ) ) {
                     // define relative path
                     CString
                        oszNot;
                     oszPath = oszFile;
                     if ( ! oszPath.mSplitFromRight ( _SYSTEM_FILESEPARATOR, oszNot ) )
                        oszPath = "";   // no path
                     else { // add file separator at end of path
                        char
                           szFileSep [ 2 ] = { _SYSTEM_FILESEPARATOR, 0 };  // construct string
                        oszPath += szFileSep;
                     }
                  }
                  // define all file name (with path)
                  CString
                     oszRead;
                  oszRead = oszPath + oszStr;
                  // try load file
                  CFileTXE
                     oFile;
                  INT
                     nErrorCode;
                  if ( ( nErrorCode = oFile.mLoad ( oszRead, *m_poScen ) ) != 1 )
                     return nErrorCode;     // error loading (1 OK, no error)
               }
            } while ( oszStr != STR_CLOSEBLOCK ); // while not close block

            printf ( MSG_CONTINUE_READING_TXE, (char*) oszFile );

            break;

         case ID_PALETTE:
            {
               // try read palette file name, if error exit
               if ( ! mRetStr ( oszStr ) ) {
                  printf ( MSG_ERROR_SPECTED, MSG_STR_FILE_NAME, STR_PALETTE );
                  return 0;
               }

               // convert 3DFC file separator '/' to system file seperator
               oszStr.mReplace ( _3DFC_FILESEPARATOR, _SYSTEM_FILESEPARATOR );

               // read palette from PCX file
               // define file name to be readed
               CString
                  oszPath;
               // verify if the name readed not have drive-separator
               if ( ! oszStr.mTestFromLeft ( DRIVESEPARATOR ) ) {
                  // define relative path
                  CString
                     oszNot;
                  oszPath = oszFile;
                  if ( ! oszPath.mSplitFromRight ( _SYSTEM_FILESEPARATOR, oszNot ) )
                     oszPath = "";   // no path
                  else { // add file separator at end of path
                     char
                        szFileSep [ 2 ] = { _SYSTEM_FILESEPARATOR, 0 };  // construct string
                     oszPath += szFileSep;
                  }
               }
               // define all file name (with path)
               CString
                  oszRead;
               oszRead = oszPath + oszStr;
               // load palette from file
               CFilePCX
                  oFile;
               CMem
                  oPal;
               printf ( MSG_READING_PALETTE, (char*) oszRead );
               if ( ! oFile.mLoadPalette ( oszRead, oPal ) ) { // try load
                  printf ( MSG_NOT_FOUND, (char*) oszRead );
                  return 0;
               }
               // try add in scenery
               if ( ! m_poScen->mAddPalette ( oPal ) )
                  return 0;
            }

            break;

         case ID_LIGHTTABLES:
            {
               // try read open block, if can't return
               if ( ! mTestNextToken ( STR_OPENBLOCK ) ) {
                  printf ( MSG_ERROR_SPECTED, STR_OPENBLOCK, STR_LIGHTTABLES );
                  return 0;
               }

               // load light table files (one at a time)
               do {
                  // try read light table identifier, if error: exit
                  if ( ! mRetStr ( oszStr ) ) {
                     printf ( MSG_ERROR_SPECTED, MSG_STR_ID, STR_LIGHTTABLES );
                     return 0;
                  }
                  // verify if not read end of block
                  if ( oszStr != STR_CLOSEBLOCK ) {
                     // define ID as a number
                     DWORD
                        dwID = (DWORD) atoi ( (char*) oszStr );

                     // try read light table file name, if error exit
                     if ( ! mRetStr ( oszStr ) ) {
                        printf ( MSG_ERROR_SPECTED, MSG_STR_FILE_NAME, STR_LIGHTTABLES );
                        return 0;
                     }

                     // convert 3DFC file separator '/' to system file seperator
                     oszStr.mReplace ( _3DFC_FILESEPARATOR, _SYSTEM_FILESEPARATOR );

                     // try define file name (load is made bellow)
                     // define file name to be readed
                     CString
                        oszPath;
                     // verify if the name readed not have drive-separator
                     if ( ! oszStr.mTestFromLeft ( DRIVESEPARATOR ) ) {
                        // define relative path
                        CString
                           oszNot;
                        oszPath = oszFile;
                        if ( ! oszPath.mSplitFromRight ( _SYSTEM_FILESEPARATOR, oszNot ) )
                           oszPath = "";   // no path
                        else { // add file separator at end of path
                           char
                              szFileSep [ 2 ] = { _SYSTEM_FILESEPARATOR, 0 };  // construct string
                           oszPath += szFileSep;
                        }
                     }
                     // define all file name (with path)
                     CString
                        oszRead;
                     oszRead = oszPath + oszStr;

                     // try read the number of light levels
                     if ( ! mRetStr ( oszStr ) ) {
                        printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_LIGHTTABLES );
                        return 0;
                     }

                     // define light levels as a number
                     DWORD
                        dwNumLevels = (INT) atoi ( (char*) oszStr );

                     // try load light table from file (if can't is
                     // regenerated by CScenery)
                     CLightTable
                        oLgT;
                     printf ( MSG_READING_LIGHTTABLE, (char*) oszRead );
                     oLgT.mCreate ( oszRead );

                     // try read the light source Red color component
                     if ( ! mRetStr ( oszStr ) ) {
                        printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_LIGHTTABLES );
                        return 0;
                     }
                     // define as a floating number
                     float
                        fNormR = (float) atof ( (char*) oszStr );

                     // try read the light source Green color component
                     if ( ! mRetStr ( oszStr ) ) {
                        printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_LIGHTTABLES );
                        return 0;
                     }
                     // define as a floating number
                     float
                        fNormG = (float) atof ( (char*) oszStr );

                     // try read the light source Blue color component
                     if ( ! mRetStr ( oszStr ) ) {
                        printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_LIGHTTABLES );
                        return 0;
                     }
                     // define as a floating number
                     float
                        fNormB = (float) atof ( (char*) oszStr );

                     // try add in scenery
                     if ( ! m_poScen->mAddLightTable ( oLgT, dwID, oszRead, dwNumLevels,
                                                       fNormR, fNormG, fNormB ) )
                        return 0;
                  }
               } while ( oszStr != STR_CLOSEBLOCK ); // while not close block
            }

            break;

         case ID_TEXTURES:
            // try read open block, if can't return
            if ( ! mTestNextToken ( STR_OPENBLOCK ) ) {
               printf ( MSG_ERROR_SPECTED, STR_OPENBLOCK, STR_TEXTURES );
               return 0;
            }

            // load textures from PCX files (one at a time)
            do {

               // try read texture identifier, if error: exit
               if ( ! mRetStr ( oszStr ) ) {
                  printf ( MSG_ERROR_SPECTED, MSG_STR_ID, STR_TEXTURES );
                  return 0;
               }
               // verify if not read end of block
               if ( oszStr != STR_CLOSEBLOCK ) {

                  // define ID as a number
                  DWORD
                     dwID = (DWORD) atoi ( (char*) oszStr );

                  // try read texture file name, if error exit
                  if ( ! mRetStr ( oszStr ) ) {
                     printf ( MSG_ERROR_SPECTED, MSG_STR_FILE_NAME, STR_TEXTURES );
                     return 0;
                  }

                  // convert 3DFC file separator '/' to system file seperator
                  oszStr.mReplace ( _3DFC_FILESEPARATOR, _SYSTEM_FILESEPARATOR );

                  // read texture file
                  // define file name to be readed
                  CString
                     oszPath;
                    // verify if the name readed not have drive-separator
                  if ( ! oszStr.mTestFromLeft ( DRIVESEPARATOR ) ) {
                     // define relative path
                     CString
                        oszNot;
                     oszPath = oszFile;
                     if ( ! oszPath.mSplitFromRight ( _SYSTEM_FILESEPARATOR, oszNot ) )
                        oszPath = "";   // no path
                     else { // add file separator at end of path
                        char
                           szFileSep [ 2 ] = { _SYSTEM_FILESEPARATOR, 0 };  // construct string
                        oszPath += szFileSep;
                     }
                  }
                  // define all file name (with path)
                  CString
                     oszRead;
                  oszRead = oszPath + oszStr;
                  // load PCX file
                  CFilePCX
                     oFile;
                  CTexture
                     oTx;
                  printf ( MSG_READING_TEXTURE, (char*) oszRead );
                  if ( ! oFile.mLoadTexture ( oszRead, oTx ) ) { // try load
                     printf ( MSG_NOT_FOUND, (char*) oszRead );
                     return 0;
                  }

                  // try add in scenery
                  if ( ! m_poScen->mAddTexture ( dwID, oTx ) )
                     return 0;
               }
            } while ( oszStr != STR_CLOSEBLOCK ); // while not close block

            break;

         case ID_START:
            {
               // try read camera identifier, if error: exit
               if ( ! mRetStr ( oszStr ) ) {
                  printf ( MSG_ERROR_SPECTED, MSG_STR_ID, STR_START );
                  return 0;
               }

               // define start instance ID to be used as viewer camera as a number
               DWORD
                  dwCameraID = (DWORD) atoi ( (char*) oszStr );

               // try read parent instance ID (inside who the camera is),
               // if error: exit
               if ( ! mRetStr ( oszStr ) ) {
                  printf ( MSG_ERROR_SPECTED, MSG_STR_ID, STR_START );
                  return 0;
               }

               // define parent instance
               DWORD
                  dwParentInstID = (DWORD) atoi ( (char*) oszStr );

               // try read sector ID (inside what sector the camera is),
               // if error: exit
               if ( ! mRetStr ( oszStr ) ) {
                  printf ( MSG_ERROR_SPECTED, MSG_STR_ID, STR_START );
                  return 0;
               }

               // define parent sector
               DWORD
                  dwParentSectID = (DWORD) atoi ( (char*) oszStr );

               // try read background color (used to clear frame buffer), if error: exit
               if ( ! mRetStr ( oszStr ) ) {
                  printf ( MSG_ERROR_SPECTED, MSG_STR_ID, STR_START );
                  return 0;
               }

               // define start instance ID to be used as viewer camera as a number
               INT
                  nClear = (DWORD) atoi ( (char*) oszStr );

               // verify if an error found
               if ( ! m_poScen->mSetStart ( dwCameraID, dwParentInstID, dwParentSectID, nClear ) )
                  return 0;
            }

            break;

         case ID_INSTANCES:
            // try read open block, if can't return
            if ( ! mTestNextToken ( STR_OPENBLOCK ) ) {
               printf ( MSG_ERROR_SPECTED, STR_OPENBLOCK, STR_INSTANCES );
               return 0;
            }

            // load identifies that must be instanceated and descriptions
            do {

               // try read intance ID, if error: exit
               if ( ! mRetStr ( oszStr ) ) {
                  printf ( MSG_ERROR_SPECTED, MSG_STR_INSTANCES, STR_INSTANCES );
                  return 0;
               }

               // verify if not read end of block
               if ( oszStr != STR_CLOSEBLOCK ) {

                  // define ID as a number
                  DWORD
                     dwInstID = (DWORD) atoi ( (char*) oszStr );

                  // try read the model (object) ID source, if error exit
                  if ( ! mRetStr ( oszStr ) ) {
                     printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_INSTANCES );
                     return 0;
                  }

                  // verify if read close block (}) -> in wrong place -> abort
                  if ( oszStr == STR_CLOSEBLOCK ) {
                     printf ( MSG_ERROR_UNSPECTED_EOB, (char*) oszFile, STR_INSTANCES );
                     return 0;
                  }

                  // convert to number
                  DWORD
                     dwObjectID = (DWORD) atoi ( (char*) oszStr );

                  // try read where place the instance (inside the sector
                  // of what object), if error exit. Here read the destin instance ID.
                  if ( ! mRetStr ( oszStr ) ) {
                     printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_INSTANCES );
                     return 0;
                  }

                  // verify if read close block (}) -> in wrong place -> abort
                  if ( oszStr == STR_CLOSEBLOCK ) {
                     printf ( MSG_ERROR_UNSPECTED_EOB, (char*) oszFile, STR_INSTANCES );
                     return 0;
                  }

                  // convert to number
                  DWORD
                     dwDestinInstID = (DWORD) atoi ( (char*) oszStr );

                  // try read the destin sector ID
                  if ( ! mRetStr ( oszStr ) ) {
                     printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_INSTANCES );
                     return 0;
                  }

                  // verify if read close block (}) -> in wrong place -> abort
                  if ( oszStr == STR_CLOSEBLOCK ) {
                     printf ( MSG_ERROR_UNSPECTED_EOB, (char*) oszFile, STR_INSTANCES );
                     return 0;
                  }

                  // convert to number
                  DWORD
                     dwDestinSectorID = (DWORD) atoi ( (char*) oszStr );

                  // read orientation and position
                  tagfXYZ
                     stI,
                     stJ,
                     stK;
                  tagXYZ
                     stPos;

                  // read object position

                  // try read X
                  if ( ! mRetStr ( oszStr ) ) {
                     printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_INSTANCES );
                     return 0;
                  }

                  // verify if read close block (}) -> in wrong place -> abort
                  if ( oszStr == STR_CLOSEBLOCK ) {
                     printf ( MSG_ERROR_UNSPECTED_EOB, (char*) oszFile, STR_INSTANCES );
                     return 0;
                  }

                  // convert to number
                  stPos.nX = (DWORD) atoi ( (char*) oszStr );

                  // try read Y
                  if ( ! mRetStr ( oszStr ) ) {
                     printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_INSTANCES );
                     return 0;
                  }

                  // verify if read close block (}) -> in wrong place -> abort
                  if ( oszStr == STR_CLOSEBLOCK ) {
                     printf ( MSG_ERROR_UNSPECTED_EOB, (char*) oszFile, STR_INSTANCES );
                     return 0;
                  }

                  // convert to number
                  stPos.nY = (DWORD) atoi ( (char*) oszStr );

                  // try read Z
                  if ( ! mRetStr ( oszStr ) ) {
                     printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_INSTANCES );
                     return 0;
                  }

                  // verify if read close block (}) -> in wrong place -> abort
                  if ( oszStr == STR_CLOSEBLOCK ) {
                     printf ( MSG_ERROR_UNSPECTED_EOB, (char*) oszFile, STR_INSTANCES );
                     return 0;
                  }

                  // convert to number
                  stPos.nZ = (DWORD) atoi ( (char*) oszStr );

                  // read base vector I 

                  // try read IX
                  if ( ! mRetStr ( oszStr ) ) {
                     printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_INSTANCES );
                     return 0;
                  }

                  // verify if read close block (}) -> in wrong place -> abort
                  if ( oszStr == STR_CLOSEBLOCK ) {
                     printf ( MSG_ERROR_UNSPECTED_EOB, (char*) oszFile, STR_INSTANCES );
                     return 0;
                  }

                  // convert to number
                  stI.fX = (float) atof ( (char*) oszStr );

                  // try read IY
                  if ( ! mRetStr ( oszStr ) ) {
                     printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_INSTANCES );
                     return 0;
                  }

                  // verify if read close block (}) -> in wrong place -> abort
                  if ( oszStr == STR_CLOSEBLOCK ) {
                     printf ( MSG_ERROR_UNSPECTED_EOB, (char*) oszFile, STR_INSTANCES );
                     return 0;
                  }

                  // convert to number
                  stI.fY = (float) atof ( (char*) oszStr );

                  // try read IZ
                  if ( ! mRetStr ( oszStr ) ) {
                     printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_INSTANCES );
                     return 0;
                  }

                  // verify if read close block (}) -> in wrong place -> abort
                  if ( oszStr == STR_CLOSEBLOCK ) {
                     printf ( MSG_ERROR_UNSPECTED_EOB, (char*) oszFile, STR_INSTANCES );
                     return 0;
                  }

                  // convert to number
                  stI.fZ = (float) atof ( (char*) oszStr );

                  // read base vector J

                  // try read JX
                  if ( ! mRetStr ( oszStr ) ) {
                     printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_INSTANCES );
                     return 0;
                  }

                  // verify if read close block (}) -> in wrong place -> abort
                  if ( oszStr == STR_CLOSEBLOCK ) {
                     printf ( MSG_ERROR_UNSPECTED_EOB, (char*) oszFile, STR_INSTANCES );
                     return 0;
                  }

                  // convert to number
                  stJ.fX = (float) atof ( (char*) oszStr );

                  // try read JY
                  if ( ! mRetStr ( oszStr ) ) {
                     printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_INSTANCES );
                     return 0;
                  }

                  // verify if read close block (}) -> in wrong place -> abort
                  if ( oszStr == STR_CLOSEBLOCK ) {
                     printf ( MSG_ERROR_UNSPECTED_EOB, (char*) oszFile, STR_INSTANCES );
                     return 0;
                  }

                  // convert to number
                  stJ.fY = (float) atof ( (char*) oszStr );

                  // try read JZ
                  if ( ! mRetStr ( oszStr ) ) {
                     printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_INSTANCES );
                     return 0;
                  }

                  // verify if read close block (}) -> in wrong place -> abort
                  if ( oszStr == STR_CLOSEBLOCK ) {
                     printf ( MSG_ERROR_UNSPECTED_EOB, (char*) oszFile, STR_INSTANCES );
                     return 0;
                  }

                  // convert to number
                  stJ.fZ = (float) atof ( (char*) oszStr );

                  // read base vector K

                  // try read KX
                  if ( ! mRetStr ( oszStr ) ) {
                     printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_INSTANCES );
                     return 0;
                  }

                  // verify if read close block (}) -> in wrong place -> abort
                  if ( oszStr == STR_CLOSEBLOCK ) {
                     printf ( MSG_ERROR_UNSPECTED_EOB, (char*) oszFile, STR_INSTANCES );
                     return 0;
                  }

                  // convert to number
                  stK.fX = (float) atof ( (char*) oszStr );

                  // try read KY
                  if ( ! mRetStr ( oszStr ) ) {
                     printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_INSTANCES );
                     return 0;
                  }

                  // verify if read close block (}) -> in wrong place -> abort
                  if ( oszStr == STR_CLOSEBLOCK ) {
                     printf ( MSG_ERROR_UNSPECTED_EOB, (char*) oszFile, STR_INSTANCES );
                     return 0;
                  }

                  // convert to number
                  stK.fY = (float) atof ( (char*) oszStr );

                  // try read KZ
                  if ( ! mRetStr ( oszStr ) ) {
                     printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_INSTANCES );
                     return 0;
                  }

                  // verify if read close block (}) -> in wrong place -> abort
                  if ( oszStr == STR_CLOSEBLOCK ) {
                     printf ( MSG_ERROR_UNSPECTED_EOB, (char*) oszFile, STR_INSTANCES );
                     return 0;
                  }

                  // convert to number
                  stK.fZ = (float) atof ( (char*) oszStr );

                  // read copy data flag

                  // try read if need copy data to instance or only refer a model
                  if ( ! mRetStr ( oszStr ) ) {
                     printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_INSTANCES );
                     return 0;
                  }

                  // verify if read close block (}) -> in wrong place -> abort
                  if ( oszStr == STR_CLOSEBLOCK ) {
                     printf ( MSG_ERROR_UNSPECTED_EOB, (char*) oszFile, STR_INSTANCES );
                     return 0;
                  }

                  // convert to number
                  BOOL
                     bCopyData = (BOOL) atoi ( (char*) oszStr );

                  // try add in scenery
                  if ( ! m_poScen->mAddInstance ( dwInstID, dwObjectID,
                                                  dwDestinInstID, dwDestinSectorID,
                                                  stPos, stI, stJ, stK, bCopyData ) )
                     return 0;
               }
            } while ( oszStr != STR_CLOSEBLOCK ); // while not close block

            break;

         case ID_OBJECT:
            {
               // try read object identifier, if error: exit
               if ( ! mRetStr ( oszStr ) ) {
                  printf ( MSG_ERROR_SPECTED, MSG_STR_ID, STR_OBJECT );
                  return 0;
               }
               // define object ID as a number
               DWORD
                  dwObjID = (DWORD) atoi ( (char*) oszStr );

               // try read open block, if can't return
               if ( ! mTestNextToken ( STR_OPENBLOCK ) ) {
                  printf ( MSG_ERROR_SPECTED, STR_OPENBLOCK, STR_OBJECT );
                  return 0;
               }

               // show object ID
               printf ( MSG_OBJECT_ID, dwObjID );
               
               // try create object in scenery
               INT
                  nCreateCode = m_poScen->mBeginAddObject ( dwObjID );
               // verify creation error
               if ( nCreateCode != 1 )
                  return nCreateCode;

               // load object declarations (blocks inside object)
               INT
                  nReturnCode = mReadBlock ( oszFile, ID_OBJECT );
               if ( nReturnCode != 1 ) {
                  printf ( MSG_ERROR_READING_BLOCK, STR_OBJECT, dwObjID );
                  return nReturnCode;
               }

               // try signals end object read
               INT
                  nEndCode = m_poScen->mEndAddObject ();
               // verify error
               if ( nEndCode != 1 )
                  return nEndCode;
            }

            break;

         case ID_CLOSEBLOCK:
            bEnd = TRUE;
            break;

         case ID_SECTOR:
            {
               // try read sector identifier, if error: exit
               if ( ! mRetStr ( oszStr ) ) {
                  printf ( MSG_ERROR_SPECTED, MSG_STR_ID, STR_SECTOR );
                  return 0;
               }
               // define sector ID as a number
               DWORD
                  dwSectorID = (DWORD) atoi ( (char*) oszStr );

               // try read open block, if can't return
               if ( ! mTestNextToken ( STR_OPENBLOCK ) ) {
                  printf ( MSG_ERROR_SPECTED, STR_OPENBLOCK, STR_SECTOR );
                  return 0;
               }

               // try create sector in object
               INT
                  nCreateCode = m_poScen->mBeginAddSector ( dwSectorID );
               // verify creation error
               if ( nCreateCode != 1 )
                  return nCreateCode;

               // show sector ID
               printf ( MSG_SECTOR_ID, dwSectorID );

               // load sector declarations (blocks inside sector)
               INT
                  nReturnCode = mReadBlock ( oszFile, ID_SECTOR );
               if ( nReturnCode != 1 ) {
                  printf ( MSG_ERROR_READING_BLOCK, STR_SECTOR, dwSectorID );
                  return nReturnCode;
               }

               // try signals end sector read
               INT
                  nEndCode = m_poScen->mEndAddSector ();
               // verify error
               if ( nEndCode != 1 )
                  return nEndCode;
            }

            break;

         case ID_ATTRIBUTES:
            // try read open block, if can't return
            if ( ! mTestNextToken ( STR_OPENBLOCK ) ) {
               printf ( MSG_ERROR_SPECTED, STR_OPENBLOCK, STR_ATTRIBUTES );
               return 0;
            }

            // load attributes from file
            do {

               // try read the attribute 1: polygon is a portal?, error: exit
               if ( ! mRetStr ( oszStr ) ) {
                  printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_ATTRIBUTES );
                  return 0;
               }

               // verify if read tuple-separator (;) in wrong place -> abort
               if ( oszStr == STR_TUPLESEPARATOR ) {
                  printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_ATTRIBUTES );
                  return 0;
               }

               // verify if not read end of block
               if ( oszStr != STR_CLOSEBLOCK ) {

                  // define condition flag as a number
                  BOOL
                     bPortal = (BOOL) atoi ( (char*) oszStr );

                  // try read attribute 2: crossable polygon (no collision detection), error exit
                  if ( ! mRetStr ( oszStr ) ) {
                     printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_ATTRIBUTES );
                     return 0;
                  }

                  // verify if read tuple-separator (;) in wrong place -> abort
                  if ( oszStr == STR_TUPLESEPARATOR ) {
                     printf ( MSG_ERROR_SPECTED_IN, MSG_STR_VALUE, STR_ATTRIBUTES );
                     return 0;
                  }

                  // define flag as a number
                  BYTE
                     yCrossable = (BYTE) atoi ( (char*) oszStr );

                  // try read attribute 3: texture ID for polygon, error exit
                  if ( ! mRetStr ( oszStr ) ) {
                     printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_ATTRIBUTES );
                     return 0;
                  }

                  // verify if read tuple-separator (;) in wrong place -> abort
                  if ( oszStr == STR_TUPLESEPARATOR ) {
                     printf ( MSG_ERROR_SPECTED_IN, MSG_STR_VALUE, STR_ATTRIBUTES );
                     return 0;
                  }

                  // define ID as a number
                  DWORD
                     dwTextureID = (DWORD) atoi ( (char*) oszStr );

                  // try read first polygon index, if error exit
                  if ( ! mRetStr ( oszStr ) ) {
                     printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_ATTRIBUTES );
                     return 0;
                  }

                  // verify if read tuple-separator (;) in wrong place -> abort
                  if ( oszStr == STR_TUPLESEPARATOR ) {
                     printf ( MSG_ERROR_SPECTED_IN, MSG_STR_VALUE, STR_ATTRIBUTES );
                     return 0;
                  }

                  // convert to number (index) - can have 2 indexes per attr
                  INT
                     nNdxPol1 = (INT) atoi ( (char*) oszStr ),
                     nNdxPol2 = nNdxPol1;

                  // try read next token, if error exit
                  if ( ! mRetStr ( oszStr ) ) {
                     printf ( MSG_ERROR_SPECTED_IN, MSG_STR_VALUE, STR_ATTRIBUTES );
                     return 0;
                  }

                  // define what have been read
                  // if not read an attribute separator, supose read trash
                  if ( oszStr != STR_TUPLESEPARATOR ) {

                     // verify if read continuator (to), if no abort
                     if ( oszStr != STR_CONTINUATOR ) {
                        printf ( MSG_ERROR_SPECTED_IN, STR_CONTINUATOR, STR_ATTRIBUTES );
                        return 0;
                     }

                     // try read second polygons index, if error exit
                     if ( ! mRetStr ( oszStr ) ) {
                        printf ( MSG_ERROR_SPECTED_IN, MSG_STR_VALUE, STR_ATTRIBUTES );
                        return 0;
                     }

                     // verify if read tuple-separator (;) in wrong place -> abort
                     if ( oszStr == STR_TUPLESEPARATOR ) {
                        printf ( MSG_ERROR_SPECTED_IN, MSG_STR_VALUE, STR_ATTRIBUTES );
                        return 0;
                     }

                     // convert to number (index of an other polygon)
                     nNdxPol2 = (INT) atoi ( (char*) oszStr );
                  }
                  // try add in scenery
                  if ( ! m_poScen->mSetPolygonsAttributes ( bPortal, yCrossable, dwTextureID, nNdxPol1, nNdxPol2 ) )
                     return 0;
               }
            } while ( oszStr != STR_CLOSEBLOCK ); // while not close block

            break;

         case ID_POLYGONSLIGHT:
            // try read open block, if can't return
            if ( ! mTestNextToken ( STR_OPENBLOCK ) ) {
               printf ( MSG_ERROR_SPECTED, STR_OPENBLOCK, STR_POLYGONSLIGHT );
               return 0;
            }

            // load attributes from file
            do {

               // try read the light value
               if ( ! mRetStr ( oszStr ) ) {
                  printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_POLYGONSLIGHT );
                  return 0;
               }

               // verify if read tuple-separator (;) in wrong place -> abort
               if ( oszStr == STR_TUPLESEPARATOR ) {
                  printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_POLYGONSLIGHT );
                  return 0;
               }

               // verify if not read end of block
               if ( oszStr != STR_CLOSEBLOCK ) {

                  // define light level as number
                  float
                     fLight = (float) atof ( (char*) oszStr );

                  // try read light table ID, if error exit
                  if ( ! mRetStr ( oszStr ) ) {
                     printf ( MSG_ERROR_SPECTED_IN, MSG_STR_VALUE, STR_POLYGONSLIGHT );
                     return 0;
                  }

                  // verify if read tuple-separator (;) in wrong place -> abort
                  if ( oszStr == STR_TUPLESEPARATOR ) {
                     printf ( MSG_ERROR_SPECTED_IN, MSG_STR_VALUE, STR_POLYGONSLIGHT );
                     return 0;
                  }

                  // convert to number
                  DWORD
                     dwLightID = (DWORD) atoi ( (char*) oszStr );

                  // try read first polygon index, if error exit
                  if ( ! mRetStr ( oszStr ) ) {
                     printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_POLYGONSLIGHT );
                     return 0;
                  }

                  // verify if read tuple-separator (;) in wrong place -> abort
                  if ( oszStr == STR_TUPLESEPARATOR ) {
                     printf ( MSG_ERROR_SPECTED_IN, MSG_STR_VALUE, STR_POLYGONSLIGHT );
                     return 0;
                  }

                  // convert to number (index) - can have 2 indexes per attr
                  INT
                     nNdxPol1 = (INT) atoi ( (char*) oszStr ),
                     nNdxPol2 = nNdxPol1;

                  // try read next token, if error exit
                  if ( ! mRetStr ( oszStr ) ) {
                     printf ( MSG_ERROR_SPECTED_IN, MSG_STR_VALUE, STR_POLYGONSLIGHT );
                     return 0;
                  }

                  // define what have been read
                  // if not read an attribute separator, supose read trash
                  if ( oszStr != STR_TUPLESEPARATOR ) {

                     // verify if read continuator (..), if no abort
                     if ( oszStr != STR_CONTINUATOR ) {
                        printf ( MSG_ERROR_SPECTED_IN, STR_CONTINUATOR, STR_POLYGONSLIGHT );
                        return 0;
                     }

                     // try read second polygons index, if error exit
                     if ( ! mRetStr ( oszStr ) ) {
                        printf ( MSG_ERROR_SPECTED_IN, MSG_STR_VALUE, STR_POLYGONSLIGHT );
                        return 0;
                     }

                     // verify if read tuple-separator (;) in wrong place -> abort
                     if ( oszStr == STR_TUPLESEPARATOR ) {
                        printf ( MSG_ERROR_SPECTED_IN, MSG_STR_VALUE, STR_POLYGONSLIGHT );
                        return 0;
                     }

                     // convert to number (index of an other polygon)
                     nNdxPol2 = (INT) atoi ( (char*) oszStr );
                  }
                  // try add in scenery
                  if ( ! m_poScen->mSetPolygonsLight ( fLight, dwLightID, nNdxPol1, nNdxPol2 ) )
                     return 0;
               }
            } while ( oszStr != STR_CLOSEBLOCK ); // while not close block

            break;

         case ID_VERTEXES:
            // try read open block, if can't return
            if ( ! mTestNextToken ( STR_OPENBLOCK ) ) {
               printf ( MSG_ERROR_SPECTED, STR_OPENBLOCK, STR_VERTEXES );
               return 0;
            }

            // load vertexes from file
            do {

               // try read X component, if error: exit
               if ( ! mRetStr ( oszStr ) ) {
                  printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_VERTEXES );
                  return 0;
               }

               // verify if not read end of block
               if ( oszStr != STR_CLOSEBLOCK ) {

                  // define X as a number
                  INT
                     nX = (INT) atoi ( (char*) oszStr );

                  // try read Y, if error exit
                  if ( ! mRetStr ( oszStr ) ) {
                     printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_VERTEXES );
                     return 0;
                  }

                  // verify if read close block (}) -> in wrong place -> abort
                  if ( oszStr == STR_CLOSEBLOCK ) {
                     printf ( MSG_ERROR_UNSPECTED_EOB, (char*) oszFile, STR_VERTEXES );
                     return 0;
                  }

                  // convert to number
                  INT
                     nY = (INT) atoi ( (char*) oszStr );

                  // try read Z, if error exit
                  if ( ! mRetStr ( oszStr ) ) {
                     printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_VERTEXES );
                     return 0;
                  }

                  // verify if read close block (}) -> in wrong place -> abort
                  if ( oszStr == STR_CLOSEBLOCK ) {
                     printf ( MSG_ERROR_UNSPECTED_EOB, (char*) oszFile, STR_VERTEXES );
                     return 0;
                  }

                  // convert to number
                  INT
                     nZ = (INT) atoi ( (char*) oszStr );
                  // try add in scenery
                  if ( ! m_poScen->mAddVertex ( nX, nY, nZ, TRUE ) )
                     return 0;
               }
            } while ( oszStr != STR_CLOSEBLOCK ); // while not close block

            break;

         case ID_PORTALS:
            // try read open block, if can't return
            if ( ! mTestNextToken ( STR_OPENBLOCK ) ) {
               printf ( MSG_ERROR_SPECTED, STR_OPENBLOCK, STR_PORTALS );
               return 0;
            }

            // load relations between polygons and sectors from file
            do {

               // try read polygon index, if error: exit
               if ( ! mRetStr ( oszStr ) ) {
                  printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_PORTALS );
                  return 0;
               }

               // verify if not read end of block
               if ( oszStr != STR_CLOSEBLOCK ) {

                  // define index as a number
                  DWORD
                     dwNdxPol = (DWORD) atoi ( (char*) oszStr );

                  // try read object identifier, if error exit
                  if ( ! mRetStr ( oszStr ) ) {
                     printf ( MSG_ERROR_SPECTED, MSG_STR_ID, STR_PORTALS );
                     return 0;
                  }

                  // verify if read close block (}) -> in wrong place -> abort
                  if ( oszStr == STR_CLOSEBLOCK ) {
                     printf ( MSG_ERROR_UNSPECTED_EOB, (char*) oszFile, STR_PORTALS );
                     return 0;
                  }

                  // convert to number
                  DWORD
                     dwObjectID = (DWORD) atoi ( (char*) oszStr );

                  // try read sector identifier, if error exit
                  if ( ! mRetStr ( oszStr ) ) {
                     printf ( MSG_ERROR_SPECTED, MSG_STR_ID, STR_PORTALS );
                     return 0;
                  }

                  // verify if read close block (}) -> in wrong place -> abort
                  if ( oszStr == STR_CLOSEBLOCK ) {
                     printf ( MSG_ERROR_UNSPECTED_EOB, (char*) oszFile, STR_PORTALS );
                     return 0;
                  }

                  // convert to number
                  DWORD
                     dwSectorID = (DWORD) atoi ( (char*) oszStr );

                  // try add in scenery
                  if ( ! m_poScen->mAddPortal ( dwNdxPol, dwObjectID, dwSectorID ) )
                     return 0;
               }
            } while ( oszStr != STR_CLOSEBLOCK ); // while not close block

            break;

         case ID_POLYGONS:
            {
               // try read open block, if can't return
               if ( ! mTestNextToken ( STR_OPENBLOCK ) ) {
                  printf ( MSG_ERROR_SPECTED, STR_OPENBLOCK, STR_POLYGONS );
                  return 0;
               }

               // list of vertexes for one polygon
               CList
                  oIndexes;

               // load polygons from file
               do {

                  // empty list
                  oIndexes.mRemoveAll ();

                  // try read the first token, if error: exit
                  if ( ! mRetStr ( oszStr ) ) {
                     printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_POLYGONS );
                     return 0;
                  }

                  // verify if read tuple-separator (;) in wrong place -> abort
                  if ( oszStr == STR_TUPLESEPARATOR ) {
                     printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_POLYGONS );
                     return 0;
                  }

                  // verify if not read end of block
                  if ( oszStr != STR_CLOSEBLOCK ) {

                     // execute while not find a tuple separator
                     do {

                        // convert to numeric value and insert in list
                        CNodeInt
                           *poVertex = new CNodeInt;
                        // verify if not alocate memory
                        if ( poVertex == NULL ) {
                           printf ( MSG_ERROR_MEMORY_LOW );
                           return 0;
                        }
                        // convert
                        poVertex->m_nValue = (INT) atoi ( (char*) oszStr );
                        // insert
                        oIndexes.mSetEnd ( poVertex );

                        // try read next token, if error exit
                        if ( ! mRetStr ( oszStr ) ) {
                           printf ( MSG_ERROR_SPECTED, MSG_STR_VALUE, STR_POLYGONS );
                           return 0;
                        }

                        // verify if read close block (}) -> in wrong place -> abort
                        if ( oszStr == STR_CLOSEBLOCK ) {
                           printf ( MSG_ERROR_UNSPECTED_EOB, (char*) oszFile, STR_POLYGONS );
                           return 0;
                        }

                     } while ( oszStr != STR_TUPLESEPARATOR ); // end of polygon indexes

                     // try add in scenery
                     if ( ! m_poScen->mAddPolygon ( oIndexes ) )
                        return 0;
                  }

               } while ( oszStr != STR_CLOSEBLOCK ); // while not close block

               // empty list
               oIndexes.mRemoveAll ();

               break;
            }
      }
   }

   return 1;
}



/* Define ID names in array (to user messages)
**********************************************************/
void CFileTXE::mDefineIDNames ()
{
   m_aoIDs [ ID_ERROR         ] = STR_ERROR;
   m_aoIDs [ ID_NOBLOCK       ] = STR_NOBLOCK;
   m_aoIDs [ ID_DATA          ] = STR_DATA;
   m_aoIDs [ ID_READ          ] = STR_READ;
   m_aoIDs [ ID_TEXTURES      ] = STR_TEXTURES;
   m_aoIDs [ ID_PALETTE       ] = STR_PALETTE;
   m_aoIDs [ ID_OBJECT        ] = STR_OBJECT;
   m_aoIDs [ ID_SECTOR        ] = STR_SECTOR;
   m_aoIDs [ ID_POLYGONS      ] = STR_POLYGONS;
   m_aoIDs [ ID_VERTEXES      ] = STR_VERTEXES;
   m_aoIDs [ ID_PORTALS       ] = STR_PORTALS;
   m_aoIDs [ ID_ATTRIBUTES    ] = STR_ATTRIBUTES;
   m_aoIDs [ ID_OPENBLOCK     ] = STR_OPENBLOCK;
   m_aoIDs [ ID_CLOSEBLOCK    ] = STR_CLOSEBLOCK;
   m_aoIDs [ ID_OPENTUPLE     ] = STR_OPENTUPLE;
   m_aoIDs [ ID_CLOSETUPLE    ] = STR_CLOSETUPLE;
   m_aoIDs [ ID_INSTANCES     ] = STR_INSTANCES;
   m_aoIDs [ ID_START         ] = STR_START;
   m_aoIDs [ ID_LIGHTTABLES   ] = STR_LIGHTTABLES;
   m_aoIDs [ ID_POLYGONSLIGHT ] = STR_POLYGONSLIGHT;
}



/* Define the type of block that the token correspond.
oszStr - the token to be tested
nParentBlock - context, the parent block
ID values can be found in header file.
Acording the context one block can't ocurr (it is an error)
Close block is one possible block type.
return ID type or ID_ERROR if block in worng place.
**********************************************************/
INT CFileTXE::mClassifyBlock ( CString oszStr, INT nParentBlock )
{
   // verify cases

   if ( oszStr == STR_READ )
      // verify context
      if ( nParentBlock == ID_NOBLOCK )
         return ID_READ;
      else
         return ID_ERROR;

   if ( oszStr == STR_INSTANCES )
      // verify context
      if ( nParentBlock == ID_NOBLOCK )
         return ID_INSTANCES;
      else
         return ID_ERROR;

   if ( oszStr == STR_START )
      // verify context
      if ( nParentBlock == ID_NOBLOCK )
         return ID_START;
      else
         return ID_ERROR;

   if ( oszStr == STR_TEXTURES )
      // verify context
      if ( nParentBlock == ID_NOBLOCK )
         return ID_TEXTURES;
      else
         return ID_ERROR;

   if ( oszStr == STR_PALETTE )
      // verify context
      if ( nParentBlock == ID_NOBLOCK )
         return ID_PALETTE;
      else
         return ID_ERROR;

   if ( oszStr == STR_LIGHTTABLES )
      // verify context
      if ( nParentBlock == ID_NOBLOCK )
         return ID_LIGHTTABLES;
      else
         return ID_ERROR;

   if ( oszStr == STR_OBJECT )
      // verify context
      if ( nParentBlock == ID_NOBLOCK )
         return ID_OBJECT;
      else
         return ID_ERROR;

   if ( oszStr == STR_SECTOR )
      // verify context
      if ( nParentBlock == ID_OBJECT )
         return ID_SECTOR;
      else
         return ID_ERROR;

   // close block is one possible name!
   if ( oszStr == STR_CLOSEBLOCK )
      // verify context
      if ( nParentBlock == ID_OBJECT || nParentBlock == ID_SECTOR )
         return ID_CLOSEBLOCK;
      else
         return ID_ERROR;

   if ( oszStr == STR_PORTALS )
      // verify context
      if ( nParentBlock == ID_OBJECT || nParentBlock == ID_SECTOR )
         return ID_PORTALS;
      else
         return ID_ERROR;

   if ( oszStr == STR_POLYGONS )
      // verify context
      if ( nParentBlock == ID_OBJECT || nParentBlock == ID_SECTOR )
         return ID_POLYGONS;
      else
         return ID_ERROR;

   if ( oszStr == STR_VERTEXES )
      // verify context
      if ( nParentBlock == ID_OBJECT || nParentBlock == ID_SECTOR )
         return ID_VERTEXES;
      else
         return ID_ERROR;

   if ( oszStr == STR_ATTRIBUTES )
      // verify context
      if ( nParentBlock == ID_OBJECT || nParentBlock == ID_SECTOR )
         return ID_ATTRIBUTES;
      else
         return ID_ERROR;

   if ( oszStr == STR_POLYGONSLIGHT )
      // verify context
      if ( nParentBlock == ID_OBJECT || nParentBlock == ID_SECTOR )
         return ID_POLYGONSLIGHT;
      else
         return ID_ERROR;

   return ID_ERROR;
}



/* Try read an token. If the token read is the same
oszSpected return TRUE else FALSE.
**********************************************************/
BOOL CFileTXE::mTestNextToken ( CString oszSpected )
{
   // try read a token
   CString
      oszStr;
   if ( ! mRetStr ( oszStr ) )
      return FALSE;

   // verify if string is the spected string
   if ( oszStr == oszSpected )
      return TRUE;

   return FALSE;
}
