/* Store a sector data: polygon mesh and instances of objects inside this sector.
A sector must define a convex group of polys in scenery.

TIPS:
- this class derive from CMesh, see CMesh for more info

WARNING:
- after change sector data call CSector::mBeginAddSector and imediatly before
  call CSector::mEndAddSector.
*/



#include <string.h>

#include "defs.h"

#include "node.h"

#include "list.h"

#include "polygon.h"

#include "nodepol.h"

#include "nodeint.h"

#include "nodepor.h"

#include "map.h"

#include "mesh.h"

#include "math.h"

#include "refsys.h"

#include "camsys.h"

#include "object.h"

#include "nodeobj.h"

#include "sector.h"



/* constructor - init empty sector
**********************************************************/
CSector::CSector ()
{
}



/* destructor - desalocate sector data, if alocated
**********************************************************/
CSector::~CSector ()
{
   mDestroy ();
}



/* Desalocate sector data, if alocated
**********************************************************/
void CSector::mDestroy ()
{
   CMesh::mDestroy ();

   m_oInstances.mRemoveAll ();
}



/* Prepare sector to reseve data.
Call after change data, before call CSector::mEndAddSector
return TRUE - ok / FALSE - error
**********************************************************/
BOOL CSector::mBeginAddSector ()
{
   return mBeginAddMesh ();
}



/* Update sector data acording definition.
Call after define all data and before use this data.
return TRUE - ok / FALSE - error
**********************************************************/
BOOL CSector::mEndAddSector ()
{
   return mEndAddMesh ();
}



/* Search in this sector for a given object instance and, if
   not found, in the objects that are inside this sector.
return the object pointer or NULL if not found.
**********************************************************/
CNodeObject* CSector::mSearchObject ( DWORD dwObjID )
{
   CNodeObject
      *poObj,
      *poObjFound;

   // scan objects vailable
   DWORD
      dwNumInstances = m_oInstances.mRetSize ();
   for ( DWORD dwcInd = 0; dwcInd < dwNumInstances; dwcInd ++ ) {

      // for each key (object) in sector
      DWORD
         dwKey = m_oInstances.mRetKey ( dwcInd );
      poObj = (CNodeObject*) m_oInstances.mRetValue ( dwKey );

      // verify if is the same key
      if ( dwKey == dwObjID )
         // return the object
         return poObj;         
      else
         // search inside the sectors on this object for the ID
         if ( ( poObjFound = poObj->mSearchObject ( dwObjID ) ) != NULL )
            return poObjFound;
   }

   // if not found
   return NULL;
}
                  
